import tokenize
from token import tok_name
import sys

from .util import Token, isrepl


stack_for_repl = None
repl_counter = None


def dowhile(a):
    global stack_for_repl
    global repl_counter

    if stack_for_repl is not None and len(stack_for_repl) > 0:
        stack = stack_for_repl
        l = stack[-1]
        #l.pop(-1) # remove previous ENDMARKER
        if repl_counter == 1:
            l.insert(0, Token(tokenize.INDENT, '    '))
        repl_counter += 1
    else:
        stack = []
        l = []
        stack_for_repl = None
        repl_counter = None

    checking_while_start_index = None
    while_condition_tokens = []

    for type, name in a:
        #print(tok_name[type], name)
        l.append(Token(type, name))
        #print(l)
        #print(stack)
        if checking_while_start_index is not None:
            if l[-1].name == ':':
                checking_while_start_index = None
            elif l[-1].name == '\n':
                dowhile_condition_tokens = [Token(tokenize.NAME, 'if'), Token(tokenize.NAME, f'_HDYTTO_DOWHILE_COUNTER_{len(stack)}'), Token(tokenize.NAME, 'and'), Token(tokenize.NAME, 'not'), Token(tokenize.OP, '(')] + l[checking_while_start_index + 1:-1] + [Token(tokenize.OP, ')'), Token(tokenize.OP, ':'), Token(tokenize.NEWLINE, '\n'), Token(tokenize.INDENT, l[8 + (2 if repl_counter is not None else 0)].name + '    '), Token(tokenize.NAME, 'break'), Token(tokenize.NEWLINE, '\n'), Token(tokenize.DEDENT, ''), Token(tokenize.NAME, f'_HDYTTO_DOWHILE_COUNTER_{len(stack)}'), Token(tokenize.VBAREQUAL, '|='), Token(tokenize.NAME, 'True'), Token(tokenize.NEWLINE, '\n')]
                for i, token in enumerate(dowhile_condition_tokens):
                    l.insert(9 + i + (2 if repl_counter is not None else 0), token)
                while len(l) > checking_while_start_index + len(dowhile_condition_tokens):
                    del l[checking_while_start_index + len(dowhile_condition_tokens)]
                checking_while_start_index = None
                if isrepl():
                    base_indent = l[10].name
                    for i in range(len(l)):
                        if i > 0 and l[i].type == tokenize.INDENT:
                            if i == 10 or l[i].name != base_indent:
                                l[i] = Token(l[i].type, '    ' + l[i].name)
                if len(stack) > 1:
                    stack[-2].extend(stack.pop(-1))
                    l = stack[-1]
                else:
                    #print('######')
                    #print(l)
                    #print('######')
                    for i, (t, n) in enumerate(l):
                        if isrepl() and i > 10 and l[i].type == tokenize.INDENT and l[i].name == base_indent:
                            continue
                        if t != tokenize.ENDMARKER:
                            yield t, n
                    l = []
                    stack = []
                    if stack_for_repl is not None:
                        yield tokenize.DEDENT, ''
                        stack_for_repl = None
                        repl_counter = None

        if len(l) <= 1:
            continue
        if l[-2].name == 'do' and l[-1].name == ':':
            l.pop(-1)
            l.pop(-1)
            l = []
            stack.append(l)
            #l.append(Token(tokenize.NAME, 'if'))
            #l.append(Token(tokenize.NAME, 'True'))
            #l.append(Token(tokenize.OP, ':'))
            #l.append(Token(tokenize.NEWLINE, '\n'))
            l.append(Token(tokenize.NAME, f'_HDYTTO_DOWHILE_COUNTER_{len(stack)}'))
            l.append(Token(tokenize.OP, '='))
            l.append(Token(tokenize.NAME, 'False'))
            l.append(Token(tokenize.NEWLINE, '\n'))
            l.append(Token(tokenize.NAME, 'while'))
            l.append(Token(tokenize.NAME, 'True'))
            l.append(Token(tokenize.OP, ':'))
            continue
        if len(stack) == 0:
            yield l.pop(0)
            continue
        if l[-1].name == 'while':
            checking_while_start_index = len(l) - 1
            continue
    if isrepl():
        if len(stack) == 0:
            pass
        elif repl_counter is None:
            stack_for_repl = stack
            repl_counter = 1
            yield tokenize.NAME, 'if'
            yield tokenize.NAME, 'True'
            yield tokenize.OP, ':'
            yield tokenize.NEWLINE, '\n'
        else:
            if l[0].type != tokenize.INDENT: # there maybe error when inputing do: multi lines, so removing the stacks
                stack_for_repl = None
                repl_counter = None
                for t, n in l:
                    yield t, n
            else:
                yield tokenize.INDENT, '    '
                yield tokenize.NAME, 'pass'
                yield tokenize.NEWLINE, '\n'
    #if len(l) > 7 and l[-1].type == tokenize.ENDMARKER and len(stack) > 0:
        # Called in REPL:
        # >>> do:
    #    stack_for_repl = stack
    #    repl_counter = 1
    #    print('hoge')
    #elif stack_for_repl is not None:
    else:
        for t, n in l:
            yield t, n
    l = []



from typing import List, Optional
from dataclasses import dataclass
import tokenize
from token import tok_name
from keyword import iskeyword
import traceback
import sys
import inspect

from .util import Token, syntax_error_util, isrepl


@dataclass
class StackItem:
    const_vars: List[str]
    indent: int
    class_name: Optional[str] = None,


def _in_stack(name, stack):
    for item in stack:
        if name in item.const_vars:
            return True

const_vars_for_repl = None
const_vars_stack_for_repl = None

def const(a):
    global const_vars_for_repl
    global const_vars_stack_for_repl

    if isrepl() and const_vars_for_repl is not None:
        const_vars_stack = const_vars_stack_for_repl
        const_vars = const_vars_stack[-1].const_vars
    else:
        const_vars = []
        const_vars_stack = [StackItem(const_vars, 0)]

    lineno = 1
    l = []
    for type, name in a:
        if isrepl():
            const_vars_for_repl = const_vars
            const_vars_stack_for_repl = const_vars_stack

        #print(tok_name[type], name)
        l.append(Token(type, name))
        #print(l)
        #print(const_vars_stack)
        if l[0].name == 'const':
            if len(l) == 1:
                continue
            if l[1].type != tokenize.NAME:
                yield l.pop(0)
                continue
            if len(l) == 2:
                continue
            if l[2].name != '=':
                syntax_error_util.raise_error('missing = in const declaration', lineno=lineno, offset=len(l[0].name) + 1 + len(l[1].name), line=l[0].name + ' ' + l[1].name)
            if l[-1].name != '\n' and l[-1].type != tokenize.ENDMARKER:
                continue
            if _in_stack(l[1].name, const_vars_stack):
                syntax_error_util.raise_error(f'redeclaration of const {l[1].name}', lineno=lineno, offset=sum([len(each_l.name) for each_l in l[:-1]]) + len(l) - 2, line=' '.join([each_l.name for each_l in l[:-1]]))
            lineno += 1
            const_vars.append(l[1].name)
            #print(const_vars)
            l.pop(0) # remove const
            while len(l) > 0:
                yield l.pop(0)
            continue
        #if l[0].name == 'class': # TODO: class variables
        #    if l[-1].type != tokenize.INDENT:
        #        continue
        #    const_vars = []
        #    const_vars_stack.append(StackItem(const_vars, 1, l[1].name))
        #    while len(l) > 0:
        #        yield l.pop(0)
        #    continue
        if l[0].name == 'def':
            if l[-1].name != ':':
                continue
            const_vars = []
            const_vars_stack.append(StackItem(const_vars, 0))
            for i in range(len(l)):
                if l[i].name == 'const':
                    const_vars.append(l[i + 1].name)
            while len(l) > 0:
                if l[0].name == 'const':
                    l.pop(0)
                    continue
                yield l.pop(0)
            continue
        if l[0].type == tokenize.INDENT:
            const_vars_stack[-1].indent += 1
            yield l.pop(0)
            continue
        if l[0].type == tokenize.DEDENT:
            const_vars_stack[-1].indent -= 1
            yield l.pop(0)
            if len(const_vars_stack) > 1 and const_vars_stack[-1].indent <= 0:
                const_vars_stack.pop(-1)
                const_vars = const_vars_stack[-1].const_vars
            continue
        if l[0].type == tokenize.NAME:
            if not _in_stack(l[0].name, const_vars_stack):
                yield l.pop(0)
                continue
            if len(l) == 1:
                continue
            if l[1].name not in ['=', '+=', '-=', '*=', '/=', '%=', '&=', '|=', '^=', '<<=', '>>=', '**=', '//=', '@=', ':=']:
                yield l.pop(0)
                continue
            if l[-1].name != '\n' and l[-1].type != tokenize.ENDMARKER:
                continue
            if isrepl() and len(const_vars_stack) > 1:
                while len(const_vars_stack) > 1:
                    const_vars_stack.pop(-1)
                    const_vars = const_vars_stack[-1].const_vars
            syntax_error_util.raise_error(f'invalid assignment to const {l[0].name}', lineno=lineno, offset=sum([len(each_l.name) for each_l in l[:-1]]) + len(l) - 2, line=' '.join([each_l.name for each_l in l[:-1]]))

        if l[0].name == '\n':
            lineno += 1
        yield l.pop(0)

    #print('end')
    for t, n in l:
        yield t, n
    l = []



import tokenize
from tokenize import TokenInfo
from token import tok_name
import codecs, encodings
from io import StringIO
from encodings import utf_8
import sys

from .increment import increment
from .comment import comment
from .dowhile import dowhile
from .const import const
from .util import Token, isrepl, syntax_error_util

UTF8 = encodings.search_function('utf8')

def recalculate_3tuples(a):
    l = []
    indent_num = 0
    indent_stack = []
    row = 1
    for type, name in a:
        l.append(Token(type, name))
        #print(l)
        if l[-1].type == tokenize.INDENT:
            indent_num += 1
            indent_stack.append(l[-1].name)
        elif l[-1].type == tokenize.DEDENT:
            if indent_num > 0:
                indent_num -= 1
                indent_stack.pop(-1)
        elif l[-1].type == tokenize.NEWLINE or l[-1].type == tokenize.NL or l[-1].type == tokenize.ENDMARKER:
            physical_line = ' '.join([n for _, n in l])
            if l[0].type != tokenize.INDENT and len(indent_stack) > 0:
                col = len(indent_stack[-1])
                physical_line = indent_stack[-1] + physical_line
            else:
                col = 0
            for t, n in l:
                yield t, n, (row, col), (row, (col := col + len(n))), physical_line
                if t != tokenize.INDENT and t != tokenize.DEDENT:
                    col += 1
            l = []
            row += 1


def check_a(a):
    for t, n, x, y, z in a:
        print(f'{t}{n} ({x},{y},{z})', end='')
        yield t, n, x, y, z

def transform(stream):
    #print("call transform")
    syntax_error_util.find_filepath(stream)
    sys.tracebacklimit = None

    stream = comment(stream)

    a = tokenize.generate_tokens(StringIO(stream).readline)

    a = increment(a)
    a = const(a)
    a = dowhile(a)
    a = recalculate_3tuples(a)
    #a = check_a(a)

    #a = list(tokenize.untokenize(a))
    #print(''.join(a))
    #return iter(a)

    return tokenize.untokenize(a)

def decode(input, errors='strict'):
    #print('call decode')
    if isinstance(input, memoryview):
        input = input.tobytes().decode('utf-8')
    input = transform(input)
    return input, len(input)

class IncrementalDecoder(utf_8.IncrementalDecoder):
    def decode(self, input, final=False):
        #print('decode in incrementaldecoder')
        return transform(super().decode(input, final))

class StreamReader(utf_8.StreamReader):
    def __init__(self, *args, **kwargs):
        codecs.StreamReader.__init__(self, *args, **kwargs)
        self.stream = StringIO(transform(self.stream))

def search_function(s):
    #print('call search_function')
    if s != 'hdytto':
        return None

    #print('hdytto is used')
    return codecs.CodecInfo(
        name='hdytto',
        encode=UTF8.encode,
        decode=decode,
        incrementalencoder=UTF8.incrementalencoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=UTF8.streamwriter)

def register_hdytto():
    codecs.register(search_function)



import tokenize
from token import tok_name
from keyword import iskeyword

from .util import Token


def increment(a):
    l = []
    for type, name, _, _, _ in a:
        #print(tok_name[type], name)
        l.append(Token(type, name))
        #print(l)
        if l[0].type == tokenize.NAME and not iskeyword(l[0].name):
            if len(l) > 1:
                if l[1].name != '+' and l[1].name != '-':
                    yield l.pop(0)
                    continue
            if len(l) < 3:
                continue
            if l[1].name == '+' and l[2].name == '+':
                yield tokenize.OP, '('
                yield tokenize.OP, '('
                yield tokenize.NAME, l[0].name
                yield tokenize.OP, ':='
                yield tokenize.NAME, l[0].name
                yield tokenize.OP, '+'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                yield tokenize.OP, '-'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                l = []
                continue
            if l[1].name == '-' and l[2].name == '-':
                yield tokenize.OP, '('
                yield tokenize.OP, '('
                yield tokenize.NAME, l[0].name
                yield tokenize.OP, ':='
                yield tokenize.NAME, l[0].name
                yield tokenize.OP, '-'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                yield tokenize.OP, '+'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                l = []
                continue
            yield l.pop(0)
            continue
        elif l[0].name == '+':
            if len(l) > 1:
                if l[1].name != '+' and l[1].name != '-':
                    yield l.pop(0)
                    continue
            if len(l) < 3:
                continue
            if l[1].name == '+' and l[2].type == tokenize.NAME and not iskeyword(l[2].name):
                yield tokenize.OP, '('
                yield tokenize.NAME, l[2].name
                yield tokenize.OP, ':='
                yield tokenize.NAME, l[2].name
                yield tokenize.OP, '+'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                l = []
                continue
            yield l.pop(0)
            continue
        elif l[0].name == '-':
            if len(l) > 1:
                if l[1].name != '+' and l[1].name != '-':
                    yield l.pop(0)
                    continue
            if len(l) < 3:
                continue
            if l[1].name == '-' and l[2].type == tokenize.NAME and not iskeyword(l[2].name):
                yield tokenize.OP, '('
                yield tokenize.NAME, l[2].name
                yield tokenize.OP, ':='
                yield tokenize.NAME, l[2].name
                yield tokenize.OP, '-'
                yield tokenize.NUMBER, '1'
                yield tokenize.OP, ')'
                l = []
                continue
            yield l.pop(0)
            continue

        yield l.pop(0)

    for t, n in l:
        yield t, n
    l = []



from typing import NamedTuple
import traceback
import sys
import os
import re

class Token(NamedTuple):
    type: int
    name: str


def isrepl():
    return hasattr(sys, 'ps1')


class SyntaxErrorUtil:
    def raise_error(self, msg, lineno, offset, line):
        try:
            raise SyntaxError(msg, (self.filepath, lineno, offset, line))
        except:
            err = sys.exc_info()
        print(''.join(traceback.format_exception(*err)[2:]), file=sys.stderr)

        if isrepl():
            sys.tracebacklimit = -1
            raise Exception
        sys.exit(2) # 2 means SyntaxError in Unix.

    def find_filepath(self, file_content):
        if isrepl():
            self.filepath = '<stdin>'
            return

        if not self._find_filepath(sys.argv[0], file_content):
            self.filepath = '<some script>'

    def _find_filepath(self, filename, file_content):
        if not os.path.exists(filename):
            return False

        with open(filename, 'r') as f:
            f_content = f.read()
            f_content = '\n' + '\n'.join(f_content.split('\n')[1:])
            if f_content == file_content:
                self.filepath = os.path.abspath(filename)
                return True
            for line in sum([tmp_line.split(';') for tmp_line in f_content.split('\n')], []):
                froms = re.findall(r'from (\w+)', line)
                imports = re.findall(r'^\s*import ([\w,\s]+)', line)
                imports = sum([each_import.replace(' ', '').split(',') for each_import in imports], [])
                modules = froms + imports
                for module in modules:
                    if self._find_filepath(module, file_content):
                        return True
syntax_error_util = SyntaxErrorUtil()



# coding: utf-8

import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()

requires = []


setuptools.setup(
    name='hdytto',
    version='0.2',
    description='Add not pythonic syntax into your python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/TRSasasusu/hdytto',
    author='kaito kishi',
    author_email='trsasasusu@gmail.com',
    license='MIT',
    keywords='shell commands alias',
    packages=setuptools.find_packages(),
    install_requires=requires,
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires=">=3.8",
)



import re
import tokenize

from .util import Token

# `comment` uses regex because indent cannot be restored from dedent generated by tokenize.
def comment(stream):
    return re.sub(r'\/\*[^\/*]*\*\/', '', stream)
